from __future__ import nested_scopes

import QUB
from QUBB import *
from Buttons import *
from qubtree import Node
from ModelBuilder import ModelTree
from math import *
from sys import *

try:
    from datetime import datetime
except:
    pass

import win32com.client
from pythoncom import CoInitialize, CoUninitialize


def WriteFileOut(fname, lst):
    f = open(fname, 'a')
    for i, x in enumerate(lst):
        if i:
            f.write('\t')
        f.write(str(x))
    f.write('\n')
    f.close()

WriteSimOut = lambda lst: WriteFileOut('diff_sim.txt', lst)
WriteTimeOut = lambda lst: WriteFileOut('diff_time.txt', lst)


def WriteDateTime(WriteFunc):
    try:
        WriteFunc([datetime.now().strftime("%d %b %Y %H:%M:%S")])
    except:
        pass


WriteSimOut([])
WriteDateTime(WriteSimOut)
WriteSimOut(['states', 'k', 'k12', 'k21', 'k+', 'k-', 'L', 'phi', 'steps (E, phi)'])

WriteTimeOut([])
WriteDateTime(WriteTimeOut)
WriteTimeOut(['states', 'k', 'Tau black', 'Amp black', 'Tau red', 'Amp red', 'steps (E, phi)'])


def interleave(*lists):
    out = []
    for i in range(len(lists[0])):
        for l in range(len(lists)):
            out.append( lists[l][i] )
    return out


excel = None
sim_row = 1
tc_row = 1

def EnsureExcel():
    global excel, book, sim_sheet, sim_row, tc_sheet, tc_row
    
    if excel:
        try:
            excel.Visible = 1
            sim_sheet.Cells(1, 1).Value = 'states'        
            tc_sheet.Cells(1, 1).Value = 'states'        
        except:
            excel = None
    if not excel:
        #CoInitialize() -- let WordReport.py do this on startup
        
        excel = win32com.client.Dispatch("Excel.Application")
        book = excel.Workbooks.Add(-4167) # XLWBatWorksheet
        book.Sheets.Add()
        sim_sheet = book.Worksheets.Item(1)
        tc_sheet = book.Worksheets.Item(2)
        # sheet.Cells.HorizontalAlignment = -4108 # 0xffffeff4 # xlHAlignCenter
        excel.Visible = 1

        sim_sheet.Cells(1, 1).Value = 'states'        
        sim_sheet.Cells(1, 2).Value = 'k'        
        sim_sheet.Cells(1, 3).Value = 'k12'        
        sim_sheet.Cells(1, 4).Value = 'k21'        
        sim_sheet.Cells(1, 5).Value = 'k+'        
        sim_sheet.Cells(1, 6).Value = 'k-'
        sim_sheet.Cells(1, 7).Value = 'L'
        sim_sheet.Cells(1, 8).Value = 'phi'
        sim_sheet.Cells(1, 9).Value = 'steps (E, phi)'
        sim_row = 2

        tc_sheet.Cells(1, 1).Value = 'states'        
        tc_sheet.Cells(1, 2).Value = 'k'        
        tc_sheet.Cells(1, 3).Value = 'Tau black'        
        tc_sheet.Cells(1, 4).Value = 'Amp black'
        tc_sheet.Cells(1, 5).Value = 'Tau red'
        tc_sheet.Cells(1, 6).Value = 'Amp red'
        tc_sheet.Cells(1, 7).Value = 'steps (E, phi)'
        tc_row = 2

def EndExcel():
    global excel, book, sim_sheet, sim_row, tc_sheet, tc_row
    try:
        del excel
        del book
        del tc_sheet
        del sim_sheet
    except:
        pass
    excel = None
    sim_row = 1
    tc_row = 1
    #CoUninitialize()

def AddRow_sim(lst=[]):
    global excel, book, sim_sheet, sim_row
    
    EnsureExcel()
    #for i, item in enumerate(lst):
    for i in range(len(lst)):
        item = lst[i]
        sim_sheet.Cells(sim_row, i+1).Value = str(item)
    sim_row = sim_row + 1

def AddRow_tc(lst=[]):
    global excel, book, tc_sheet, tc_row
    
    EnsureExcel()
    #for i, item in enumerate(lst):
    for i in range(len(lst)):
        item = lst[i]
        tc_sheet.Cells(tc_row, i+1).Value = str(item)
    tc_row = tc_row + 1

def AddRow():
    AddRow_sim()
    AddRow_tc()


baseEstep = []
basePhistep = []


def PromptSteps(n):
    shortfall = (n-1) - len(baseEstep)
    baseEstep.extend( [0.0] * shortfall )
    basePhistep.extend( [0.5] * shortfall )
    fields = []
    for i in range(n-1):
        fields.append(DialogText('E%d'%i, 'E step %d'%(i+1), str(baseEstep[i])))
        fields.append(DialogText('phi%d'%i, 'phi step %d'%(i+1), str(basePhistep[i])))
    vals = QUB.Dialog('Enter steps', ['OK'], fields)
    for i in range(n-1):
        try:
            baseEstep[i] = float(vals['E%d'%i])
        except:
            pass
        try:
            basePhistep[i] = float(vals['phi%d'%i])
        except:
            pass


def DiffusionModel(props):
    states = listStates(props['states'])
    n = len(states)
    k = props['k']

    modelObj = QUB.ModelView.currModel
    if not modelObj:
        QUB.Report('no model')
        return

    modelTree = ModelTree()
    state = modelTree.addState( states[n-1] )
    state['x'].data = 85.0
    state['y'].data = 50.0
    
    for i in range(1,n):
        step_ix = (n-1) - i
        state = modelTree.addStateFrom(i-1, states[step_ix])
        state['x'].data = 85.0 - 70 * i / (n-1)
        state['y'].data = 50.0
        rates = modelTree.rates[i-1]['k0'].data
        rates[0] = k * pow( exp(Estep[step_ix]), phistep[step_ix] - 1 )
        rates[1] = k * pow( exp(Estep[step_ix]), phistep[step_ix] )

    modelObj.setAsNode( modelTree.node )


def DiffusionRates_Solve(props, modelObj):
    DiffusionModel(props)

    states = listStates(props['states'])
    n = len(states)
    nBlack, nRed = 0, 0
    for state in states:
        if state == 0:
            nBlack = nBlack + 1
        elif state == 1:
            nRed = nRed + 1

    if props['Mode'] == 'solve':    
        Run( 'Simulation', 'Sim' )

        k12, k21 = props['k12'], props['k21']

        modelTree = ModelTree()
        state = modelTree.addState(0)
        state['x'].data = 30.0
        state['y'].data = 50.0
        state = modelTree.addStateFrom(0, 1)
        state['x'].data = 70.0
        state['y'].data = 50.0
        rates = modelTree.rates[0]['k0'].data
        rates[0] = k12 * 0.1
        rates[1] = k21 * 0.1

        modelObj.setAsNode( modelTree.node )
        milPrefs = Node()
        milPrefs['Mode'].data = 'optimize'
        Run( 'Modeling', 'MIL', milPrefs )

        rates = modelObj.getAsNode()['Rates']['Rate']['k0'].data
        L = (rates[0] / rates[1])
        try:
            fee = log(rates[0] / k12) / log((rates[0] / k12) * (k21 / rates[1]))
            row = [props['states'], props['k'], k12, k21, rates[0], rates[1], L, fee]
            row = row + interleave(Estep[:n-1], phistep[:n-1])
            err = QUB.ResultsView.currResults.node.find('ErrorCode').data
            if len(err) and err[0]:
                row.append('Error: %d' % err[0])
            WriteSimOut(row)
        except:
            pass
        
    else: # 'time constants'
        #try:
            evalOnly = Node()
            evalOnly['Mode'].data = 'evaluate'
            Run('Modeling', 'MIL', evalOnly)
            row = [props['states'], props['k']]
            tc = QUB.ResultsView.currResults.node.find('DataSet').find('HistogramGroup').find('TimeConstants')
            row = row + [tc['Tau'].data[nBlack-1], tc['Amp'].data[nBlack-1]]
            alltc = [' '] + interleave(tc['Tau'].data, tc['Amp'].data)
            tc = tc.next('TimeConstants')
            row = row + [tc['Tau'].data[nRed-1], tc['Amp'].data[nRed-1]]
            alltc = alltc + [' '] + interleave(tc['Tau'].data, tc['Amp'].data)
            row = row + interleave(Estep[:n-1], phistep[:n-1]) + alltc
            WriteTimeOut(row)
        #except:
        #    pass

        
def floatrange(a, b, step):
    r = []
    while a < b:
        r.append(a)
        a = a + step
    return r


def DiffusionRates(props):
    global Estep, phistep

    # EnsureExcel()    

    modelObj = QUB.ModelView.currModel
    if not modelObj:
        QUB.Report('no model')
        return

    states = listStates(props['states'])
    n = len(states)
    PromptSteps(n)

    WriteOut = WriteSimOut
    if props['Mode'] == 'time constants':
        WriteOut = WriteTimeOut
    
    if (props['Mode'] == 'solve') or (props['Mode'] == 'time constants'):
        runCount = 0
        if props['stepEstep'] == 'yes':
            phistep = basePhistep[:]
            EstepPlusMinus = props['stepEstepPlusMinus']
            EstepStep = props['stepEstepStep']
            for i in range(n-1):
                Estep = baseEstep[:]
                for thisEstep in floatrange(baseEstep[i] - EstepPlusMinus, baseEstep[i] + EstepPlusMinus + EstepStep, EstepStep):
                    Estep[i] = thisEstep
                    DiffusionRates_Solve(props, modelObj)
                    runCount = runCount + 1
                WriteOut([])

        if props['stepPhistep'] == 'yes':
            Estep = baseEstep[:]
            # if this Estep is 0 then skip this phistep
            phistepPlusMinus = props['stepPhistepPlusMinus']
            phistepStep = props['stepPhistepStep']
            for i in range(n-1):
                if Estep[i] != 0.0:
                    phistep = basePhistep[:]
                    for thisPhistep in floatrange(basePhistep[i] - phistepPlusMinus, basePhistep[i] + phistepPlusMinus + phistepStep, phistepStep):
                        phistep[i] = thisPhistep
                        DiffusionRates_Solve(props, modelObj)
                        runCount = runCount + 1
                    WriteOut([])

        if runCount == 0:
            Estep = baseEstep[:]
            phistep = basePhistep[:]
            DiffusionRates_Solve(props, modelObj)
    else:
        Estep = baseEstep[:]
        phistep = basePhistep[:]
        DiffusionModel(props)

    # EndExcel()        

def acceptStates(s):
    s = s.lower()
    for c in s:
        if (c != 'c') and (c != 'o'):
            raise FieldAcceptException(s, "must be a string of C and O")
    if len(s) < 2:
        raise FieldAcceptException(s, "must have at least two states")
    
    return s

def listStates(states):
    lst = []
    for c in states:
        if c == 'c':
            lst.append(0)
        elif c == 'o':
            lst.append(1)
    return lst
    

AddTaskButton( 'Difsn Rates', # button text
               DiffusionRates,   # function
               'Difsn',       # short text for small buttons
               0,            # show settings dialog on left-click?
               1,            # show settings dialog on right-click?
               "Ask Tony Auerbach",   # tool-tip
               "Ask Tony Auerbach",   # description in "Edit Buttons"
               "Diffusion rates properties:", # caption for the settings dialog
               [TextField("states", "States:", "CCCOO", acceptStates),
                TextField("k", "k:", "10000", acceptFloat),
                TextField("k12", "k12:", "10000", acceptFloat),
                TextField("k21", "k21:", "10000", acceptFloat),
                ChoiceField("Mode", "Mode:", ['create model', 'solve', 'time constants'], 'solve'),
                ChoiceField("stepEstep", "Step Estep:", ['yes', 'no'], 'no'),
                TextField("stepEstepPlusMinus", " within +/-:", "3", acceptFloat),
                TextField("stepEstepStep", " step by:", "1", acceptFloat),
                ChoiceField("stepPhistep", "Step phiStep:", ['yes', 'no'], 'no'),
                TextField("stepPhistepPlusMinus", "  within +/-:", ".5", acceptFloat),
                TextField("stepPhistepStep", "  step by:", ".2", acceptFloat)],
               'Modeling' )

